IMPLEMENTATION MODULE Person;

        (********************************************************)
        (*                                                      *)
        (*      Web Family Tree: module to look up and          *)
        (*           display the data for a person              *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            6 July 2001                     *)
        (*  Last edited:        19 May 2003                     *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings;

FROM OurTypes IMPORT
    (* type *)  IDString, DataString, FilenameString, Line;

FROM GELookup IMPORT
    (* type *)  Database, RecordTree,
    (* proc *)  OpenDatabase, CloseDatabase,  CharSetOf,
                SeekToMatchingID, LoadRecord, DiscardTree, CopyTree,
                GetName, GetField, ExtractSubrecord, DisplayRawLines,
                GetDateRange,
                WritePersonLink, WritePersonLink2, WriteDatabaseName,
                StartReading, LoadNextINDIRecord, NonEmpty,
                SavePosition, RestorePosition;

FROM FileOps IMPORT
    (* type *)  FilePos;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM LowLevel IMPORT
    (* proc *)  EVAL;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST Nul = CHR(0);

TYPE
    (* Information about one person.                       *)
    (*      ID      the ID in the GEDCOM file              *)
    (*      DB      the database we're looking up          *)
    (*      Tree    result of looking up the GEDCOM file   *)

    PersonData = POINTER TO RECORD
                     ID: IDString;
                     DB: Database;
                     Tree: RecordTree;
                 END (*RECORD*);

    (* Records in the 'to do' list of those procedures that replace     *)
    (* recursion by a job list.                                         *)

    JobPtr = POINTER TO JobRecord;
    JobRecord = RECORD
                    next: JobPtr;
                    DB: Database;
                    ID: IDString;
                    level: CARDINAL;
                    prefix: DataString;
                    tree: RecordTree;
                END (*RECORD*);

    (* Date and place of some event. *)

    DatePlace = RECORD
                    empty: BOOLEAN;
                    date: DataString;
                    place: DataString;
                END (*RECORD*);

(************************************************************************)

PROCEDURE GetPersonData (DatabaseName: FilenameString;  PersonID: IDString;
                                       VAR (*OUT*) result: PersonData): BOOLEAN;

    (* Looks up the database for person with identifier PersonID, fills *)
    (* the result record with the information.                          *)

    BEGIN
        NEW (result);
        result^.ID := PersonID;
        IF NOT OpenDatabase (result^.DB, DatabaseName) THEN
            DISPOSE (result);
            RETURN FALSE;
        END (*IF*);
        SeekToMatchingID (result^.DB, PersonID);
        LoadRecord (result^.DB, result^.Tree);
        RETURN TRUE;
    END GetPersonData;

(**********************************************************************)

PROCEDURE OurCharSet (data: PersonData;
                      VAR (*OUT*) CharSetName: DataString);

    (* Returns the character encoding used by this database.  *)

    BEGIN
        IF data = NIL THEN
            CharSetName[0] := Nul;
        ELSE
            CharSetOf (data^.DB, CharSetName);
        END (*IF*);
    END OurCharSet;

(**********************************************************************)

PROCEDURE DiscardPersonData (VAR (*INOUT*) data: PersonData);

    (* Disposes of the PersonData information. *)

    BEGIN
        IF data <> NIL THEN
            CloseDatabase (data^.DB);
            DiscardTree (data^.Tree);
            DISPOSE (data);
        END (*IF*);
    END DiscardPersonData;

(************************************************************************)

PROCEDURE ExtractDateAndPlace (VAR (*INOUT*) info: RecordTree;
                               VAR (*OUT*) result: DatePlace);

    (* On entry, info is supposed to contain date and place information.*)
    (* We extract this to result, and dispose of the tree.              *)

    BEGIN
        WITH result DO
            empty := TRUE;
            date[0] := Nul;
            place[0] := Nul;
        END (*WITH*);
        EVAL (GetField ('DATE', info, result.date));
        EVAL (GetField ('PLAC', info, result.place));
        DiscardTree (info);
        WITH result DO
            empty := (date[0] = Nul) AND (place[0] = Nul);
        END (*WITH*);
    END ExtractDateAndPlace;

(***********************************************************************)

PROCEDURE SendLine (line: ARRAY OF CHAR);

    (* Sends a string, terminated with CRLF, to standard output. *)

    BEGIN
        WriteString (line);  WriteLn;
    END SendLine;

(************************************************************************)

PROCEDURE WrDateAndPlace (DP: DatePlace);

    (* Writes a date and place to standard output.  *)

    BEGIN
        IF DP.empty THEN
            WriteString ("(not recorded)");
        ELSE
            IF DP.date[0] = Nul THEN
                WriteString ("(date not recorded)");
            ELSE
                WriteString (DP.date);
            END (*IF*);
            IF DP.place[0] <> Nul THEN
                WriteString ("<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
                WriteString (DP.place);
            END (*IF*);
        END (*IF*);
    END WrDateAndPlace;

(************************************************************************)

PROCEDURE ExtractTag (string: ARRAY OF CHAR;  VAR (*INOUT*) DB: Database;
                                              VAR (*OUT*) tag: IDString);

    (* Extracts the substring of string enclosed by @ delimiters.  *)

    VAR pos1, pos2: CARDINAL;  found: BOOLEAN;
        database: DataString;

    BEGIN

        Strings.FindNext ('@', string, 0, found, pos1);
        IF found THEN
            Strings.FindNext ('@', string, pos1+1, found, pos2);
            IF NOT found THEN
                pos2 := LENGTH (string);
            END (*IF*);
            Strings.Extract (string, pos1+1, pos2-pos1-1, tag);
        ELSE
            tag := "";
        END (*IF*);

        (* Now check whether the tag is of the form database:id *)

        Strings.FindNext (':', tag, 0, found, pos1);
        IF found AND (pos1 > 0) THEN
            Strings.Extract (tag, 0, pos1, database);
            Strings.Extract (tag, pos1+1, LENGTH(tag)-pos1-1, tag);
            EVAL (OpenDatabase (DB, database));
        END (*IF*);

    END ExtractTag;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Writes a decimal number to standard output. *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)

PROCEDURE WriteMarriage (M: CARDINAL;  DB: Database;  FamID: IDString);

    (* Writes the details of marriage M. *)

    VAR T, subtree: RecordTree;  PersonID: IDString;
        string: DataString;
        newDB: Database;
        DP: DatePlace;
        NeedBreak: BOOLEAN;

    BEGIN
        WriteString ("<h3>Family ");
        IF M > 1 THEN
            WriteCard (M);
        END (*IF*);
        WriteString ("</h3>");
        NeedBreak := FALSE;

        SeekToMatchingID (DB, FamID);
        LoadRecord (DB, T);

        (* Date and place of marriage. *)

        IF ExtractSubrecord ('MARR', T, string, subtree) THEN
            IF NonEmpty(subtree) THEN
                ExtractDateAndPlace (subtree, DP);
                IF NOT DP.empty THEN
                    WriteString ("Date: ");
                    WrDateAndPlace (DP);
                    WriteLn;
                    NeedBreak := TRUE;
                END (*IF*);
                DiscardTree (subtree);
            END (*IF*);
        END (*IF*);

        (* Husband *)

        IF GetField ('HUSB', T, string) THEN
            newDB := DB;
            ExtractTag (string, newDB, PersonID);
            IF NeedBreak THEN
                WriteString ("<br>");
            END (*IF*);
            WriteString ("Husband: ");
            WritePersonLink (newDB, PersonID);
            WriteLn;
            NeedBreak := TRUE;
        END (*IF*);

        (* Wife *)

        IF GetField ('WIFE', T, string) THEN
            newDB := DB;
            ExtractTag (string, newDB, PersonID);
            IF NeedBreak THEN
                WriteString ("<br>");
            END (*IF*);
            WriteString ("Wife: &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
            WritePersonLink (newDB, PersonID);
            WriteLn;
            NeedBreak := TRUE;
        END (*IF*);

        (* Children *)

        IF GetField ('CHIL', T, string) THEN
            IF NeedBreak THEN
                WriteString ("<p>");
            END (*IF*);
            WriteString ("Children:");
            WriteLn;
            REPEAT
                WriteString ("<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
                newDB := DB;
                ExtractTag (string, newDB, PersonID);
                WritePersonLink (newDB, PersonID);
                WriteLn;
            UNTIL NOT GetField ('CHIL', T, string);
        END (*IF*);

        (* Notes. *)

        IF ExtractSubrecord ('NOTE', T, string, subtree)
                 AND (string[0] <> Nul) THEN
            WriteString ("<h4>Marriage notes</h4>");
            WriteLn;
            WriteString (string);
            WHILE GetField ('CONT', subtree, string) DO
                WriteChar (' ');
                WriteString (string);
            END (*IF*);
            WriteLn;
            DiscardTree (subtree);
        END (*IF*);

        (* Miscellaneous details. *)

        IF NonEmpty(T) THEN
            WriteString ("<h4>Other marriage details</h4>");
            DisplayRawLines (T);
        END (*IF*);

        DiscardTree (T);

    END WriteMarriage;

(************************************************************************)

PROCEDURE WriteParents (DB: Database;  FamID: IDString);

    (* Writes the names of the two partners to the given family ID,     *)
    (* labelled as father and mother.                                   *)

    VAR T: RecordTree;  FatherID, MotherID: IDString;
        string: DataString;
        newDB: Database;

    BEGIN
        SeekToMatchingID (DB, FamID);
        LoadRecord (DB, T);
        IF GetField ('HUSB', T, string) THEN
            newDB := DB;
            ExtractTag (string, newDB, FatherID);
            WriteString ("<p>Father: &nbsp;");
            WritePersonLink (newDB, FatherID);
        END (*IF*);
        IF GetField ('WIFE', T, string) THEN
            newDB := DB;
            ExtractTag (string, newDB, MotherID);
            WriteString ("<br>Mother: ");
            WritePersonLink (newDB, MotherID);
        END (*IF*);
        DiscardTree (T);
    END WriteParents;

(************************************************************************)

PROCEDURE DisplayPersonInfo (DB: Database;  VAR (*INOUT*) T: RecordTree;
                             prefix: ARRAY OF CHAR);

    (* Displays all the information for one person. *)

    VAR datum: DataString;  subtree: RecordTree;
        ID: IDString;  M: CARDINAL;
        newDB: Database;
        DP: DatePlace;

    BEGIN
        (* Use name and date range as heading. *)

        WriteString ("<h2>");
        IF prefix[0] <> Nul THEN
            WriteString ('<font color=RED>');
            WriteString (prefix);
            WriteString ('</font>');
            WriteString ('&nbsp;&nbsp;&nbsp;');
        END (*IF*);
        IF NOT GetName (T, datum) THEN
            datum := "No name";
        END (*IF*);
        WriteString (datum);
        GetDateRange (T, datum);
        IF datum[0] <> Nul THEN
            WriteString ("&nbsp;&nbsp;");  WriteString (datum);
        END (*IF*);
        SendLine ("</h2>");

        (* Sex. *)

        IF GetField ('SEX', T, datum) THEN
            WriteString ("Sex: ");
            WriteString (datum);
            WriteLn;
        END (*IF*);

        (* Date and place of birth. *)

        IF ExtractSubrecord ('BIRT', T, datum, subtree) THEN
            IF NonEmpty(subtree) THEN
                ExtractDateAndPlace (subtree, DP);
                IF NOT DP.empty THEN
                    WriteString ("<P>Born: ");
                    WrDateAndPlace (DP);
                    WriteLn;
                END (*IF*);
                DiscardTree (subtree);
            END (*IF*);
        END (*IF*);

        (* Date and place of death. *)

        IF ExtractSubrecord ('DEAT', T, datum, subtree) THEN
            IF NonEmpty(subtree) THEN
                ExtractDateAndPlace (subtree, DP);
                IF NOT DP.empty THEN
                    WriteString ("<P>Died: ");
                    WrDateAndPlace (DP);
                    WriteLn;
                END (*IF*);
                DiscardTree (subtree);
            END (*IF*);
        END (*IF*);

        (* Parents. *)

        IF GetField ('FAMC', T, datum) THEN
            newDB := DB;
            ExtractTag (datum, newDB, ID);
            WriteParents (newDB, ID);
            WriteLn;
        END (*IF*);

        (* Marriages. *)

        M := 1;
        WHILE GetField ('FAMS', T, datum) DO
            newDB := DB;
            ExtractTag (datum, newDB, ID);
            WriteMarriage (M, newDB, ID);
            INC (M);
        END (*WHILE*);

        (* Notes. *)

        IF ExtractSubrecord ('NOTE', T, datum, subtree)
                            AND (datum[0] <> Nul) THEN
            WriteString ("<h3>Notes</h3>");
            WriteLn;
            REPEAT
                SendLine ('<p>');
                WriteString (datum);
                WHILE GetField ('CONT', subtree, datum) DO
                    WriteChar (' ');
                    WriteString (datum);
                END (*IF*);
                WriteLn;
                DiscardTree (subtree);
                SendLine ('</p>');
            UNTIL NOT ExtractSubrecord ('NOTE', T, datum, subtree);
        END (*IF*);

        (* Display what's left in raw form. *)

        IF NonEmpty(T) THEN
            WriteString ("<h3>Other details</h3>");
            DisplayRawLines (T);
            DiscardTree (T);
        END (*IF*);

    END DisplayPersonInfo;

(************************************************************************)

PROCEDURE MakeButton (DB: Database;  ID: IDString;
                      option, label: ARRAY OF CHAR);

    (* Puts a button on the web page. *)

    BEGIN
        WriteString ('<form method="post" action="wft.cmd?D=');
        WriteDatabaseName (DB);
        WriteString (';P=');
        WriteString (ID);
        WriteString (';V=');
        WriteString (option);
        SendLine ('">');
        SendLine ('<input type=hidden name="dummy" value="">');
        WriteString ('<input type="submit" value="');
        WriteString (label);
        WriteString ('">');
        SendLine ('</form>');
    END MakeButton;

(************************************************************************)

PROCEDURE DisplayPerson (data: PersonData);

    (* Displays the information in the PersonData record. *)

    BEGIN
        (* The option buttons. *)

        WriteString ('<TABLE BORDER=0 CELLPADDING=10 CELLSPACING=0 RULES="ROWS" >');
        WriteString ('<TR ALIGN=CENTER VALIGN=TOP>');

        WriteString ('<TD>');
        MakeButton (data^.DB, data^.ID, 'D', 'Descendants');
        WriteString ('</TD>');

        WriteString ('<TD>');
        MakeButton (data^.DB, data^.ID, 'A', 'Ancestors');
        WriteString ('</TD>');

        WriteString ('<TD>');
        MakeButton (data^.DB, data^.ID, 'E', 'Everyone');
        WriteString ('</TD>');

        SendLine ('</TR></TABLE>');

        DisplayPersonInfo (data^.DB, data^.Tree, '');

    END DisplayPerson;

(************************************************************************)

PROCEDURE AppendCard (N: CARDINAL;  VAR (*INOUT*) str: ARRAY OF CHAR);

    (* Converts N to decimal, appends it to str. *)

    VAR j: CARDINAL;

    BEGIN
        IF N > 9 THEN
            AppendCard (N DIV 10, str);
            N := N MOD 10;
        END (*IF*);
        j := LENGTH(str);
        str[j] := CHR(ORD('0') + N);
        str[j+1] := Nul;
    END AppendCard;

(************************************************************************)

PROCEDURE DisplayDescendants (data: PersonData;  ShowDetails: BOOLEAN);

    (* Displays the names of this person and all direct ancestors. *)

    VAR ToDo, InsertionPoint: JobPtr;

    (********************************************************************)

    PROCEDURE AddJob (DB: Database;  ID: IDString;  level: CARDINAL;
                      prefix: DataString;  Info: RecordTree);

        (* Adds one record into the list of jobs to be done.  The       *)
        (* new record is inserted after the one identified as           *)
        (* InsertionPoint, and then InsertionPoint is changed to point  *)
        (* to the record just inserted.                                 *)

        VAR this: JobPtr;

        BEGIN
            NEW (this);
            this^.DB := DB;
            this^.ID := ID;
            this^.level := level;
            this^.prefix := prefix;
            this^.tree := Info;
            this^.next := NIL;
            IF InsertionPoint = NIL THEN
                this^.next := ToDo;
                ToDo := this;
            ELSE
                this^.next := InsertionPoint^.next;
                InsertionPoint^.next := this;
            END (*IF*);
            InsertionPoint := this;
        END AddJob;

    (********************************************************************)

    VAR this: JobPtr;
        j, M, N, level: CARDINAL;
        datum, prefix, newprefix: DataString;
        ID, FamilyID: IDString;
        Info: RecordTree;
        T2, TChild, TFamily: RecordTree;
        DB, famsDB, newDB: Database;

    BEGIN
        IF ShowDetails THEN
            MakeButton (data^.DB, data^.ID, 'D', 'Names only');
        ELSE
            MakeButton (data^.DB, data^.ID, 'D+', 'Complete details (long)');
        END (*IF*);
        SendLine ("<h2>Direct descendants</h2>");

        (* Start by putting a single job on the ToDo list. *)

        ToDo := NIL;  InsertionPoint := NIL;
        AddJob (data^.DB, data^.ID, 1, "1", data^.Tree);

        (* Note that we have moved the 'Tree' structure to an entry in  *)
        (* the ToDo list, from which it will eventually be discarded.   *)
        (* Thus, the caller must be informed that this tree has been    *)
        (* deleted.                                                     *)

        data^.Tree := NIL;

        (* Now deal with the jobs on the ToDo list.  Initially there is *)
        (* only one such job, but as we proceed we will be adding other *)
        (* jobs.  It would have been easier to do this operation via a  *)
        (* recursive solution, but the recursive approach could cause a *)
        (* stack overflow, which is why I've moved to the notion of     *)
        (* having a list of jobs that are waiting to be done.           *)

        WHILE ToDo <> NIL DO

            this := ToDo;
            ToDo := this^.next;
            InsertionPoint := NIL;
            DB := this^.DB;
            level := this^.level;
            prefix := this^.prefix;
            ID := this^.ID;
            Info := this^.tree;
            DISPOSE (this);

            (* Write the link for this person. *)

            WriteString ("<br>");

            IF ShowDetails THEN
                WriteString ("<hr>");
                T2 := CopyTree (Info);
                DisplayPersonInfo (DB, T2, prefix);
                DiscardTree (T2);
                WriteString ("<br>");
                WriteLn;
            ELSE
                FOR j := 1 TO level DO
                    WriteString ("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
                END (*FOR*);
                WriteString (prefix);
                WriteString ("&nbsp;&nbsp;&nbsp;");
                WriteLn;
                WritePersonLink2 (DB, ID, TRUE, Info);
                WriteLn;
            END (*IF*);
            Strings.Append ('.', prefix);
            N := 1;

            (* For each marriage ... *)

            M := 1;
            WHILE GetField ('FAMS', Info, datum) DO
                famsDB := DB;
                ExtractTag (datum, famsDB, FamilyID);

                SeekToMatchingID (famsDB, FamilyID);
                LoadRecord (famsDB, TFamily);

                (* Extract the children of this marriage, and add *)
                (* them to the list of pending jobs.              *)

                WHILE GetField ('CHIL', TFamily, datum) DO
                    newprefix := prefix;
                    AppendCard (N, newprefix);
                    newDB := famsDB;
                    ExtractTag (datum, newDB, ID);
                    SeekToMatchingID (newDB, ID);
                    LoadRecord (newDB, TChild);
                    AddJob (newDB, ID, level+1, newprefix, TChild);
                    INC (N);
                END (*WHILE*);

                DiscardTree (TFamily);
                INC (M);

            END (*WHILE*);

            DiscardTree (Info);

        END (*WHILE*);

    END DisplayDescendants;

(************************************************************************)

PROCEDURE DisplayAncestors (data: PersonData;  ShowDetails: BOOLEAN);

    (* Displays the names of this person and all direct ancestors. *)

    VAR ToDo: JobPtr;

    (********************************************************************)

    PROCEDURE AddJob (DB: Database;  ID: IDString;  level: CARDINAL;
                      prefix: DataString;  Info: RecordTree);

        (* Adds one record to the head of the list of jobs to be done. *)

        VAR this: JobPtr;

        BEGIN
            NEW (this);
            this^.DB := DB;
            this^.ID := ID;
            this^.level := level;
            this^.prefix := prefix;
            this^.tree := Info;
            this^.next := ToDo;
            ToDo := this;
        END AddJob;

    (********************************************************************)

    VAR TFamily: RecordTree;
        prefix: DataString;
        level: CARDINAL;

    (********************************************************************)

    PROCEDURE AddParent (DB: Database;  FieldName, SexCode: ARRAY OF CHAR);

        (* Adds a husband or wife of TFamily to the head of the ToDo list. *)

        VAR TNewPerson: RecordTree;
            PersonID: IDString;
            string, newprefix: DataString;
            newDB: Database;

        BEGIN
            IF GetField (FieldName, TFamily, string) THEN
                newprefix := prefix;
                Strings.Append (SexCode, newprefix);
                newDB := DB;
                ExtractTag (string, newDB, PersonID);
                SeekToMatchingID (newDB, PersonID);
                LoadRecord (newDB, TNewPerson);
                AddJob (newDB, PersonID, level+1, newprefix, TNewPerson);
                WriteLn;
            END (*IF*);
        END AddParent;

    (********************************************************************)

    VAR this: JobPtr;
        j: CARDINAL;
        datum: DataString;
        ID, FamilyID: IDString;
        Info, T2: RecordTree;
        DB, famcDB: Database;

    BEGIN
        IF ShowDetails THEN
            MakeButton (data^.DB, data^.ID, 'A', 'Names only');
        ELSE
            MakeButton (data^.DB, data^.ID, 'A+', 'Complete details (long)');
        END (*IF*);
        SendLine ("<h2>Ancestors in the direct line</h2>");

        (* Start by putting a single job on the ToDo list. *)

        IF NOT GetField ('SEX', data^.Tree, prefix) THEN
           prefix := '?';
        END (*IF*);
        ToDo := NIL;
        AddJob (data^.DB, data^.ID, 1, prefix, data^.Tree);

        (* Note that we have moved the 'Tree' structure to an entry in  *)
        (* the ToDo list, from which it will eventually be discarded.   *)
        (* Thus, the caller must be informed that this tree has been    *)
        (* deleted.                                                     *)

        data^.Tree := NIL;

        (* Now deal with the jobs on the ToDo list.  Initially there is *)
        (* only one such job, but as we proceed we will be adding other *)
        (* jobs.  It would have been easier to do this operation via a  *)
        (* recursive solution, but the recursive approach could cause a *)
        (* stack overflow, which is why I've moved to the notion of     *)
        (* having a list of jobs that are waiting to be done.           *)

        WHILE ToDo <> NIL DO

            this := ToDo;
            ToDo := this^.next;
            DB := this^.DB;
            level := this^.level;
            prefix := this^.prefix;
            ID := this^.ID;
            Info := this^.tree;
            DISPOSE (this);

            (* Write the link for this person (or the details, if       *)
            (* ShowDetails is TRUE.                                     *)

            WriteString ("<br>");
            IF ShowDetails THEN
                WriteString ("<hr>");
                T2 := CopyTree (Info);
                DisplayPersonInfo (DB, T2, prefix);
                DiscardTree (T2);
                WriteLn;
            ELSE
                FOR j := 1 TO level DO
                    WriteString ("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
                END (*FOR*);
                WriteString (prefix);
                WriteString ("&nbsp;&nbsp;&nbsp;");
                WriteLn;
                WritePersonLink2 (DB, ID, TRUE, Info);
                WriteLn;
            END (*IF*);

            (* Look up the family, then get husband and wife. *)

            IF GetField ('FAMC', Info, datum) THEN
                famcDB := DB;
                ExtractTag (datum, famcDB, FamilyID);

                SeekToMatchingID (famcDB, FamilyID);
                LoadRecord (famcDB, TFamily);

                (* Add the two parents to the 'ToDo' list. *)

                AddParent (famcDB, 'WIFE', 'F');
                AddParent (famcDB, 'HUSB', 'M');

                DiscardTree (TFamily);

            END (*IF*);

            DiscardTree (Info);

        END (*WHILE*);

    END DisplayAncestors;

(************************************************************************)

PROCEDURE DisplayEveryone (dummy: PersonData;  ShowDetails: BOOLEAN);

    (* Displays the names of everyone in the database.  The person      *)
    (* mentioned in the argument is unimportant, but we need some       *)
    (* person record so that we can identify the database.              *)

    VAR NextLevel, count: CARDINAL;  Lookahead: Line;
        T: RecordTree;  DB: Database;  success: BOOLEAN;
        ID: IDString;  pos: FilePos;

    BEGIN
        DB := dummy^.DB;
        IF NOT ShowDetails THEN
            MakeButton (DB, dummy^.ID, 'E+', 'Complete details (long)');
        END (*IF*);
        SendLine ("<h2>All individuals in the database</h2>");
        StartReading (DB, NextLevel, Lookahead);
        count := 0;
        REPEAT
            success := LoadNextINDIRecord (DB, ID, T, NextLevel, Lookahead);
            IF success THEN
                INC (count);
                IF ShowDetails THEN
                    pos := SavePosition (DB);
                    DisplayPersonInfo (DB, T, '');
                    RestorePosition (DB, pos);
                    WriteString ("<br><hr>");
                    WriteLn;
                ELSE
                    WriteString ("<br>");
                    WritePersonLink2 (DB, ID, TRUE, T);
                    WriteLn;
                END (*IF*);
                DiscardTree (T);
            END (*IF*);
        UNTIL NOT success;
        WriteString ("<p>Total: ");  WriteCard (count);
        WriteLn;
    END DisplayEveryone;

(************************************************************************)
(*                           INITIALISATION                             *)
(************************************************************************)

END Person.

